from importlib.metadata import PackageNotFoundError, version
import xarray as xr
from pathlib import Path

# Set xarray attributes collapsed by default
xr.set_options(display_expand_attrs=False)

# Set cache directory for ethology package
ETHOLOGY_CACHE_DIR = Path.home() / ".ethology"
ETHOLOGY_CACHE_DIR.mkdir(parents=True, exist_ok=True)

try:
    __version__ = version("ethology")
except PackageNotFoundError:
    # package is not installed
    pass



"""Utility functions for JSON schema files."""

import json
from pathlib import Path

import jsonschema


def _get_default_schema(schema_name: str) -> dict:
    """Get the default VIA or COCO schema as a dictionary."""
    schema_path = (
        Path(__file__).parent / "schemas" / f"{schema_name}_schema.json"
    )
    with open(schema_path) as file:
        schema_dict = json.load(file)
    return schema_dict


def _check_file_is_json(filepath: Path):
    """Check the input file can be read as a JSON."""
    try:
        with open(filepath) as file:
            json.load(file)
    except json.JSONDecodeError as decode_error:
        # We override the error message for clarity
        raise ValueError(
            f"Error decoding JSON data from file: {filepath}. "
            "The data being deserialized is not a valid JSON. "
        ) from decode_error


def _check_file_matches_schema(filepath: Path, schema: dict | None):
    """Check the input JSON file matches the given schema.

    The schema validation only checks the type for each specified
    key if the key exists. It does not check that the keys in the
    schema are present in the JSON file.
    """
    # Read json file
    with open(filepath) as file:
        data = json.load(file)

    # Check against schema if provided
    if schema:
        jsonschema.validate(instance=data, schema=schema)


def _check_required_properties_keys(
    required_properties_keys: list, schema: dict
):
    """Check the input schema includes the required "properties" keys."""
    # Get keys of "properties" dictionaries in schema
    properties_keys_in_schema = _extract_properties_keys(schema)

    # Get list of "properties" keys that are required but not in schema
    missing_keys = set(required_properties_keys) - set(
        properties_keys_in_schema
    )

    # Raise error if there are missing keys in the schema
    if missing_keys:
        raise ValueError(
            f"Required key(s) {sorted(missing_keys)} not found in schema."
        )


def _check_required_keys_in_dict(
    list_required_keys: list[str],
    data: dict,
    additional_message: str = "",
):
    """Check if the required keys are present in the input dictionary.

    It checks that the required keys are defined in the input dictionary.
    If they are defined, and they are dictionaries or lists,
    it additionally checks that they are not empty.

    The additional_message parameter is used to provide additional context in
    the error message for the missing keys check.
    """
    missing_keys = set(list_required_keys) - set(data.keys())
    if missing_keys:
        raise ValueError(
            f"Required key(s) {sorted(missing_keys)} not "
            f"found{additional_message}."
        )
    else:
        keys_with_empty_values = [
            key
            for key in list_required_keys
            if isinstance(data[key], dict | list) and not data[key]
        ]
        if keys_with_empty_values:
            raise ValueError(
                f"Empty value(s) found for the required key(s) "
                f"{sorted(keys_with_empty_values)}."
            )


def _extract_properties_keys(input_schema: dict, prefix: str = "") -> list:
    """Extract keys from all "properties" subdictionaries in a JSON schema.

    Recursively extract the keys of all subdictionaries in the input
    dictionary that are values to a "properties" key. The input dictionary
    represents a JSON schema dictionary
    (see https://json-schema.org/understanding-json-schema/about). The output
    is a sorted list of strings with full paths (e.g. 'parent/child').

    The "properties" key always appears as part of a set of dictionary keys
    with at least another key being "type" or "item". We use this to find the
    relevant subdictionaries.

    """
    result: list[str] = []

    # Skip if "type" key is missing in the schema
    if "type" not in input_schema:
        return result

    # If the input dictionary has a "properties" key: extract keys
    # and recurse into nested dictionaries
    if "properties" in input_schema:
        for key, value in input_schema["properties"].items():
            full_key = f"{prefix}/{key}" if prefix else key
            result.append(full_key)
            # Recurse into nested dictionaries to look for more "properties"
            # dicts
            result.extend(_extract_properties_keys(value, full_key))

    # If dictionary has "additionalProperties" key: recurse into it
    if "additionalProperties" in input_schema:
        result.extend(
            _extract_properties_keys(
                input_schema["additionalProperties"],
                prefix,
            )
        )

    # If dictionary has "items" key: recurse into it
    if "items" in input_schema:
        result.extend(
            _extract_properties_keys(
                input_schema["items"],
                prefix,
            )
        )

    # Return sorted list of keys with full paths
    return sorted(result)



"""Validators for detection datasets."""

from attrs import define, field

from ethology.validators.utils import ValidDataset


@define
class ValidBboxDetectionsDataset(ValidDataset):
    """Class for valid ``ethology`` bounding box detections datasets.

    This class validates that the input dataset:

    - is an xarray Dataset,
    - has ``image_id``, ``space``, ``id`` as dimensions,
    - has ``position``, ``shape`` and ``confidence`` as data variables,
    - ``position`` and ``shape`` span at least the dimensions ``image_id``,
      ``space`` and ``id``,
    - ``confidence`` spans at least the dimensions ``image_id`` and ``id``.


    Attributes
    ----------
    dataset : xarray.Dataset
        The xarray dataset to validate.
    required_dims : set
        The set of required dimension names: ``image_id``, ``space`` and
        ``id``.
    required_data_vars : dict[str, set]
        A dictionary mapping data variable names to their required minimum
        dimensions:

        - ``position`` maps to ``image_id``, ``space`` and ``id``,
        - ``shape`` maps to ``image_id``, ``space`` and ``id``,
        - ``confidence`` maps to ``image_id`` and ``id``.

    Raises
    ------
    TypeError
        If the input is not an xarray Dataset.
    ValueError
        If the dataset is missing required data variables or dimensions,
        or if any required dimensions are missing for any data variable.

    Notes
    -----
    The dataset can have other data variables and dimensions, but only the
    required ones are checked.

    """

    # Minimum requirements for a bbox dataset holding detections
    required_dims: set = field(
        default={"image_id", "space", "id"},
        init=False,
    )
    required_data_vars: dict = field(
        default={
            "position": {"image_id", "space", "id"},
            "shape": {"image_id", "space", "id"},
            "confidence": {"image_id", "id"},
        },
        init=False,
    )



"""Utils for validating `ethology` objects."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps

import xarray as xr
from attrs import define, field


@define
class ValidDataset(ABC):
    """An abstract base class for valid ``ethology`` datasets.

    This class validates that the input dataset:

    - is an xarray Dataset
    - contains all required dimensions
    - contains all required data variables
    - has the correct dimensions for each data variable

    Subclasses must define ``required_dims`` and ``required_data_vars``
    attributes.

    Attributes
    ----------
    dataset : xarray.Dataset
        The xarray dataset to validate.
    required_dims : set[str]
        A set of required dimension names. This attribute should be
        defined by any subclass inheriting from this class.
    required_data_vars : dict[str, set]
        A dictionary mapping data variable names to their required dimensions.
        This attribute should be defined by any subclass inheriting from
        this class.

    Raises
    ------
    TypeError
        If the input is not an xarray Dataset.
    ValueError
        If the dataset is missing required data variables or dimensions,
        or if any required dimensions are missing for any data variable.

    Notes
    -----
    The dataset can have other data variables and dimensions, but only the
    required ones are checked.

    """

    dataset: xr.Dataset = field()

    # Subclasses should override these abstract properties
    @property
    @abstractmethod
    def required_dims(self) -> set:
        """Subclasses must provide a ``required_dims`` property."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def required_data_vars(self) -> dict[str, set]:
        """Subclasses must provide a ``required_data_vars`` property."""
        pass  # pragma: no cover

    # Validators
    @dataset.validator
    def _check_dataset_type(self, attribute, value):
        """Ensure the input is an xarray Dataset."""
        if not isinstance(value, xr.Dataset):
            raise TypeError(
                f"Expected an xarray Dataset, but got {type(value)}."
            )

    @dataset.validator
    def _check_required_data_variables(self, attribute, value):
        """Ensure the dataset has all required data variables."""
        missing_vars = self.required_data_vars.keys() - set(value.data_vars)
        if missing_vars:
            raise ValueError(
                f"Missing required data variables: {sorted(missing_vars)}"
            )

    @dataset.validator
    def _check_required_dimensions(self, attribute, value):
        """Ensure the dataset has all required dimensions."""
        missing_dims = self.required_dims - set(value.dims)
        if missing_dims:
            raise ValueError(
                f"Missing required dimensions: {sorted(missing_dims)}"
            )

    @dataset.validator
    def _check_dimensions_per_data_variable(self, attribute, value):
        """Ensure the dataset has all required dimensions."""
        error_messages = []
        for data_var, dims_per_data_var in self.required_data_vars.items():
            missing_dims = dims_per_data_var - set(
                value.data_vars[data_var].coords
            )
            if missing_dims:
                error_messages.append(
                    f"data variable '{data_var}' is missing "
                    f"dimensions {sorted(missing_dims)}"
                )

        if error_messages:
            raise ValueError(
                "Some data variables are missing required dimensions:\n  - "
                + "\n  - ".join(error_messages)
            )


def _check_output(validator: type):
    """Return a decorator that validates the output of a function."""

    def decorator(function: Callable) -> Callable:
        @wraps(function)  # to preserve function metadata
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            validator(result)
            return result

        return wrapper

    return decorator


def _check_input(validator: type, input_index: int = 0):
    """Return a decorator that validates a specific input of a function.

    By default, the first input is validated. If the input index is
    larger than the number of inputs, no validation is performed.
    """

    def decorator(function: Callable) -> Callable:
        @wraps(function)
        def wrapper(*args, **kwargs):
            if len(args) > input_index:
                validator(args[input_index])
            result = function(*args, **kwargs)
            return result

        return wrapper

    return decorator



"""Validators for annotation files and datasets."""

import json
from pathlib import Path

import pandas as pd
import pandera.pandas as pa
from attrs import define, field
from pandera.typing import Index

from ethology.validators.json_schemas.utils import (
    _check_file_is_json,
    _check_file_matches_schema,
    _check_required_keys_in_dict,
    _get_default_schema,
)
from ethology.validators.utils import ValidDataset


@define
class ValidVIA:
    """Class for valid VIA JSON annotation files.

    It checks the input file is a valid JSON file, matches
    the VIA schema and contains the required keys.


    Attributes
    ----------
    path : Path | str
        Path to the VIA JSON file, passed as an input.
    schema : dict
        The JSON schema is set to the default VIA schema.
    required_keys : dict
        The required keys for the VIA JSON file.

    Raises
    ------
    ValueError
        If the JSON file cannot be decoded.
    jsonschema.exceptions.ValidationError
        If the type of any of the keys in the JSON file
        does not match the type specified in the schema.
    jsonschema.exceptions.SchemaError
        If the schema is invalid.
    ValueError
        If the VIA JSON file is missing any of the required keys.

    """

    path: Path = field(converter=Path)
    schema: dict = field(
        default=_get_default_schema("VIA"),
        init=False,
    )
    required_keys: dict = field(
        default={
            "main": ["_via_img_metadata", "_via_attributes"],
            "images": ["filename"],
            "regions": ["shape_attributes"],
            "shape_attributes": ["x", "y", "width", "height"],
        },
        init=False,
        # with init=False the attribute is always initialized
        # with the default value
    )

    # Note: the validators are applied in order
    @path.validator
    def _file_is_json(self, attribute, value):
        _check_file_is_json(value)

    @path.validator
    def _file_matches_JSON_schema(self, attribute, value):
        _check_file_matches_schema(value, self.schema)

    @path.validator
    def _file_contains_required_keys(self, attribute, value):
        """Ensure that the VIA JSON file contains the required keys."""
        # Read data as dict
        with open(value) as file:
            data = json.load(file)

        # Check first level keys
        _check_required_keys_in_dict(self.required_keys["main"], data)

        # Check keys in nested dicts
        for img_str, img_dict in data["_via_img_metadata"].items():
            # Check keys for each image dictionary
            _check_required_keys_in_dict(
                self.required_keys["images"],
                img_dict,
                additional_message=f" for {img_str}",
            )

            # Check keys for each region in an image
            for i, region in enumerate(img_dict["regions"]):
                # Check keys under first level per region
                _check_required_keys_in_dict(
                    self.required_keys["regions"],
                    region,
                    additional_message=f" for region {i} under {img_str}",
                )

                # Check keys under "shape_attributes" per region
                _check_required_keys_in_dict(
                    self.required_keys["shape_attributes"],
                    region["shape_attributes"],
                    additional_message=f" for region {i} under {img_str}",
                )


@define
class ValidCOCO:
    """Class for valid COCO JSON annotation files.

    It checks the input file is a valid JSON file, matches
    the COCO schema and contains the required keys.

    Attributes
    ----------
    path : Path | str
        Path to the COCO JSON file, passed as an input.
    schema : dict
        The JSON schema is set to the default COCO schema.
    required_keys : dict
        The required keys for the COCO JSON file.

    Raises
    ------
    ValueError
        If the JSON file cannot be decoded.
    jsonschema.exceptions.ValidationError
        If the type of any of the keys in the JSON file
        does not match the type specified in the schema.
    jsonschema.exceptions.SchemaError
        If the schema is invalid.
    ValueError
        If the COCO JSON file is missing any of the required keys.

    """

    path: Path = field(converter=Path)
    schema: dict = field(
        default=_get_default_schema("COCO"),
        init=False,
        # with init=False the attribute is always initialized
        # with the default value
    )

    # The keys of "required_keys" match the 1st level keys in a COCO JSON file
    required_keys: dict = field(
        default={
            "main": ["images", "annotations", "categories"],
            "images": ["id", "file_name", "width", "height"],
            "annotations": ["id", "image_id", "bbox", "category_id"],
            "categories": ["id", "name"],  # exclude "supercategory"
        },
        init=False,
    )

    # Note: the validators are applied in order
    @path.validator
    def _file_is_json(self, attribute, value):
        _check_file_is_json(value)

    @path.validator
    def _file_matches_JSON_schema(self, attribute, value):
        _check_file_matches_schema(value, self.schema)

    @path.validator
    def _file_contains_required_keys(self, attribute, value):
        """Ensure that the COCO JSON file contains the required keys."""

        # Helper function to singularise the input key for the
        # error message
        def _singularise_err_msg(key):
            return key[:-1] if key != "categories" else key[:-3] + "y"

        # Read file as dict
        with open(value) as file:
            data = json.load(file)

        # Check first level keys
        _check_required_keys_in_dict(self.required_keys["main"], data)

        # Check keys in every dict listed under the "images", "annotations"
        # and "categories" keys
        for ky in list(self.required_keys.keys())[1:]:
            for instance_dict in data[ky]:
                _check_required_keys_in_dict(
                    self.required_keys[ky],
                    instance_dict,
                    additional_message=(
                        f" for {_singularise_err_msg(ky)} {instance_dict}"
                    ),
                )

    @path.validator
    def _file_contains_unique_image_IDs(self, attribute, value):
        """Ensure that the COCO JSON file contains unique image IDs.

        When exporting to COCO format, the VIA tool attempts to extract the
        image ID from the image filename using ``parseInt``. As a result, if
        two or more images have the same number-based filename, the image IDs
        can be non-unique (i.e., more image filenames than image IDs). This is
        probably a bug in the VIA tool, but we need to check for this issue.
        """
        with open(value) as file:
            data = json.load(file)

        # Get number of elements in "images" list
        n_images = len(data["images"])

        # Get the image IDs
        unique_image_ids = set([img["id"] for img in data["images"]])

        # Check for duplicate image IDs
        if n_images != len(unique_image_ids):
            raise ValueError(
                "The image IDs in the input COCO file are not unique. "
                f"There are {n_images} image entries, but only "
                f"{len(unique_image_ids)} unique image IDs."
            )


@define
class ValidBboxAnnotationsDataset(ValidDataset):
    """Class for valid ``ethology`` bounding box annotations datasets.

    This class validates that the input dataset:

    - is an xarray Dataset,
    - has ``image_id``, ``space``, ``id`` as dimensions,
    - has ``position`` and ``shape`` as data variables,
    - both data variables span at least the dimensions ``image_id``,
      ``space`` and ``id``.


    Attributes
    ----------
    dataset : xarray.Dataset
        The xarray dataset to validate.
    required_dims : set[str]
        The set of required dimension names: ``image_id``, ``space`` and
        ``id``.
    required_data_vars : dict[str, set]
        A dictionary mapping data variable names to their required minimum
        dimensions:

        - ``position`` maps to ``image_id``, ``space`` and ``id``,
        - ``shape`` maps to ``image_id``, ``space`` and ``id``.

    Raises
    ------
    TypeError
        If the input is not an xarray Dataset.
    ValueError
        If the dataset is missing required data variables or dimensions,
        or if any required dimensions are missing for any data variable.

    Notes
    -----
    The dataset can have other data variables and dimensions, but only the
    required ones are checked.

    """

    # Minimum requirements for a bbox dataset holding detections
    required_dims: set = field(
        default={"image_id", "space", "id"},
        init=False,
    )
    required_data_vars: dict = field(
        default={
            "position": {"image_id", "space", "id"},
            "shape": {"image_id", "space", "id"},
        },
        init=False,
    )


class ValidBboxAnnotationsDataFrame(pa.DataFrameModel):
    """Class for valid bounding boxes intermediate dataframes.

    We use this dataframe internally as an intermediate step in the process of
    converting an input bounding box annotations file (VIA or COCO) to
    an ``ethology`` dataset. The validation checks all required columns
    exist and their types are correct.

    Attributes
    ----------
    image_filename : str
        Name of the image file.
    image_id : int
        Unique identifier for each of the images.
    image_width : int
        Width of each of the images, in the same units as the input file
        (usually pixels).
    image_height : int
        Height of each of the images, in the same units as the input file
        (usually pixels).
    x_min : float
        Minimum x-coordinate of the bounding box, in the same units as
        the input file.
    y_min : float
        Minimum y-coordinate of the bounding box, in the same units as
        the input file.
    width : float
        Width of the bounding box, in the same units as the input file.
    height : float
        Height of the bounding box, in the same units as the input file.
    category_id : int
        Unique identifier for the category, as specified in the input file.
        A value of 0 is usually reserved for the background class.
    category : str
        Category of the annotation as a string.
    supercategory : str
        Supercategory of the annotation as a string.

    Raises
    ------
    pa.errors.SchemaError
        If the input dataframe does not match the schema.

    See Also
    --------
    :class:`pandera.api.pandas.model.DataFrameModel`

    """

    # image columns
    image_filename: str = pa.Field(description="Name of the image file.")
    image_id: int = pa.Field(
        description="Unique identifier for each of the images."
    )
    image_width: int = pa.Field(
        description="Width of each of the images, "
        "in the same units as the input file (usually pixels)."
        # if not defined, it should be set to 0 in the df
    )
    image_height: int = pa.Field(
        description="Height of each of the images, "
        "in the same units as the input file (usually pixels)."
        # if not defined, it should be set to 0 in the df
    )

    # bbox columns
    x_min: float = pa.Field(
        description=(
            "Minimum x-coordinate of the bounding box, "
            "in the same units as the input file."
        )
    )
    y_min: float = pa.Field(
        description=(
            "Minimum y-coordinate of the bounding box, "
            "in the same units as the input file."
        )
    )
    width: float = pa.Field(
        description=(
            "Width of the bounding box, in the same units as the input file."
        )
    )
    height: float = pa.Field(
        description=(
            "Height of the bounding box, in the same units as the input file."
        )
    )

    # category columns
    # - always defined in COCO files exported with VIA tool
    # - optionally defined in VIA files exported with VIA tool
    category_id: int = pa.Field(
        description=(
            "Unique identifier for the category, "
            "as specified in the input file. A value of 0 "
            "is usually reserved for the background class."
        )
    )
    category: str = pa.Field(
        description="Category of the annotation as a string."
    )
    supercategory: str = pa.Field(
        description="Supercategory of the annotation as a string."
    )

    @staticmethod
    def get_empty_values() -> dict:
        """Get the default empty values for selected dataframe columns.

        The columns are those that can be undefined in VIA and COCO files:
        ``category``, ``supercategory``, ``category_id``, ``image_width`` and
        ``image_height``.

        Returns
        -------
        dict
            A dictionary with the default empty values the specified columns.

        """
        return {
            "category": "",  # it can be undefined in VIA files
            "supercategory": "",  # it can be undefined in VIA and COCO files
            "category_id": -1,  # it can be undefined in VIA files
            "image_width": 0,  # it can be undefined in VIA files
            "image_height": 0,  # it can be undefined in VIA files
        }


class ValidBboxAnnotationsCOCO(pa.DataFrameModel):
    """Class for COCO-exportable bounding box annotations dataframes.

    The validation checks the required columns exist and their types are
    correct. It additionally checks that the index and the
    ``annotation_id`` column are equal.

    Attributes
    ----------
    idx : Index[int]
        Index of the dataframe. Should be greater than or equal to 0 and equal
        to the ``annotation_id`` column.
    annotation_id : int
        Unique identifier for the annotation. Should be equal to the index.
    image_id : int
        Unique identifier for each of the images.
    image_filename : str
        Filename of the image.
    image_width : int
        Width of each of the images.
    image_height : int
        Height of each of the images.
    bbox : list[float]
        Bounding box coordinates as xmin, ymin, width, height.
    area : float
        Bounding box area.
    segmentation : list[list[float]]
        Bounding box segmentation masks as list of lists of coordinates.
    category : str
        Category of the annotation.
    supercategory : str
        Supercategory of the annotation.
    iscrowd : int
        Whether the annotation is a crowd. Should be 0 or 1.

    Raises
    ------
    pa.errors.SchemaError
        If the dataframe does not match the schema.

    Notes
    -----
    See `COCO format documentation <https://cocodataset.org/#format-data>`_
    for more details.

    See Also
    --------
    :class:`pandera.api.pandas.model.DataFrameModel`

    """

    # index
    idx: Index[int] = pa.Field(ge=0, check_name=False)

    # annotation_id
    annotation_id: int = pa.Field(
        description="Unique identifier for the annotation (index)",
    )

    # image columns
    image_id: int = pa.Field(
        description="Unique identifier for the image",
    )
    image_filename: str = pa.Field(
        description="Filename of the image",
    )
    image_width: int = pa.Field(
        description="Width of the image", ge=0, nullable=True
    )
    image_height: int = pa.Field(
        description="Height of the image", ge=0, nullable=True
    )

    # bbox data
    bbox: list[float] = pa.Field(
        description="Bounding box coordinates as xmin, ymin, width, height"
    )
    area: float = pa.Field(
        description="Bounding box area",
        ge=0,
    )
    segmentation: list[list[float]] = pa.Field(
        description="Bounding box segmentation as list of lists of coordinates"
    )

    # category columns
    # we do not require supercategories to be present in the
    # dataframe since they are not currently added to the xarray dataset
    category: str = pa.Field(
        description="Category of the annotation",
        nullable=True,
    )

    # other
    iscrowd: int = pa.Field(
        description="Whether the annotation is a crowd",
        isin=[0, 1],
        nullable=True,
    )

    @staticmethod
    def map_df_columns_to_COCO_fields() -> dict:
        """Map COCO-exportable dataframe columns to COCO fields.

        Returns
        -------
        dict
            A dictionary mapping each column in the COCO-exportable dataframe
            to the corresponding fields in the equivalent COCO file.

        """
        return {
            "images": {
                "image_id": "id",
                "image_filename": "file_name",
                "image_width": "width",
                "image_height": "height",
            },
            "categories": {
                "category_id": "id",
                "category": "name",
                "supercategory": "supercategory",
            },
            "annotations": {
                "annotation_id": "id",
                "area": "area",
                "bbox": "bbox",
                "image_id": "image_id",
                "category_id": "category_id",
                "iscrowd": "iscrowd",
                "segmentation": "segmentation",
            },
        }

    @pa.dataframe_check
    def check_idx_and_annotation_id(cls, df: pd.DataFrame) -> bool:
        """Check that the index and the ``annotation_id`` column are equal.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to check.

        Returns
        -------
        bool
            A boolean indicating whether the index and the
            ``annotation_id`` column are equal for all rows.

        """
        return all(df.index == df["annotation_id"])



"""Generate the API index page for all ``ethology`` modules."""

import os
from pathlib import Path

# Modules to exclude from the API index
exclude_modules = ["ethology.io.annotations.json_schemas"]

# Set the current working directory to the directory of this script
script_dir = Path(__file__).resolve().parent
os.chdir(script_dir)


def make_api_index():
    """Create a doctree of all ``ethology`` modules."""
    doctree = "\n"
    api_path = Path("../ethology")
    for path in sorted(api_path.rglob("*.py")):
        if path.name.startswith("_"):
            continue

        # Convert file path to module name
        rel_path = path.relative_to(api_path.parent)
        module_name = str(rel_path.with_suffix("")).replace(os.sep, ".")

        # Split the module name into parts
        # (e.g., ["ethology", "annotations", ...])
        module_parts = module_name.split(".")

        # Check against each excluded module's path
        exclude = False
        for excluded in exclude_modules:
            excluded_parts = excluded.split(".")

            # Only exclude if the module is a submodule of the excluded path
            if (
                len(module_parts) >= len(excluded_parts)
                and module_parts[: len(excluded_parts)] == excluded_parts
            ):
                exclude = True
                break
        if exclude:
            continue
        doctree += f"    {module_name}\n"

    # Get the header
    api_head_path = Path("source") / "_templates" / "api_index_head.rst"
    api_head = api_head_path.read_text()

    # Write api_index.rst with header + doctree
    output_path = Path("source") / "api_index.rst"
    with output_path.open("w") as f:
        f.write(api_head)
        f.write(doctree)


if __name__ == "__main__":
    make_api_index()



"""Load bounding boxes annotations into ``ethology``."""

import json
from pathlib import Path
from typing import Any, Literal

import numpy as np
import pandas as pd
import pandera.pandas as pa
import xarray as xr
from pandera.typing.pandas import DataFrame

from ethology.validators.annotations import (
    ValidBboxAnnotationsDataFrame,
    ValidBboxAnnotationsDataset,
    ValidCOCO,
    ValidVIA,
)
from ethology.validators.utils import _check_output


@_check_output(ValidBboxAnnotationsDataset)
def from_files(
    file_paths: Path | str | list[Path | str],
    format: Literal["VIA", "COCO"],
    images_dirs: Path | str | list[Path | str] | None = None,
) -> xr.Dataset:
    """Load an ``ethology`` bounding box annotations dataset from a file.

    Parameters
    ----------
    file_paths : pathlib.Path | str | list[pathlib.Path | str]
        Path or list of paths to the input annotation files.
    format : {"VIA", "COCO"}
        Format of the input annotation files.
    images_dirs : pathlib.Path | str | list[pathlib.Path | str], optional
        Path or list of paths to the directories containing the images the
        annotations refer to. The paths are added to the dataset
        attributes.

    Returns
    -------
    xarray.Dataset

        A valid bounding box annotations dataset with dimensions
        `image_id`, `space`, `id`, and the following arrays:

        - `position`, with dimensions (image_id, space, id),
        - `shape`, with dimensions (image_id, space, id),
        - `category`, with dimensions (image_id, id) - optional,
        - `image_shape`, with dimensions (image_id, space) - optional.

        The `category` array, if present, holds category IDs as 1-based
        integers, matching the category IDs in the input file. The dataset
        attributes include:

        - `annotation_files`: a list of paths to the input annotation files
        - `annotation_format`: the format of the input annotation files
        - `map_category_to_str`: a map from category ID to category name
        - `map_image_id_to_filename`: a map from image ID to image filename
        - `images_directories`: directory paths for the images (optional)


    Notes
    -----
    The `image_id` is assigned based on the alphabetically sorted list of
    unique image filenames across all input files. So if two images have
    the same filename but are in different input annotation files, they will
    be assigned the same image ID and their annotations will be merged.

    The `id` dimension corresponds to the annotation ID per image. It
    ranges from 0 to the maximum number of annotations per image in
    the dataset. Note that the annotation IDs are not necessarily
    consistent across images. This means that the annotations with ID=3
    in image `t` and image `t+1` will likely not correspond to the same
    individual.

    The `space` dimension holds the "x" and "y" coordinates.

    Note that supercategories are not currently added to the xarray dataset,
    even if specified in the input file.

    Examples
    --------
    Load annotations from a single COCO file:

    >>> from ethology.io.annotations import load_bboxes
    >>> ds = load_bboxes.from_files(
    ...     file_paths="path/to/annotation_file.json", format="COCO"
    ... )

    Load annotations from a single COCO file and specify the images directory:

    >>> from ethology.io.annotations import load_bboxes
    >>> ds = load_bboxes.from_files(
    ...     file_paths="path/to/annotation_file.json",
    ...     format="COCO",
    ...     images_dirs="path/to/images_dir",
    ... )

    Load annotations from two VIA files and specify multiple image directories:

    >>> from ethology.io.annotations import load_bboxes
    >>> ds = load_bboxes.from_files(
    ...     file_paths=[
    ...         "path/to/annotation_file_1.json",
    ...         "path/to/annotation_file_2.json",
    ...     ],
    ...     format="VIA",
    ...     images_dirs=["path/to/images_dir_1", "path/to/images_dir_2"],
    ... )

    """
    # Compute intermediate dataframe df
    if isinstance(file_paths, list):
        df_all = _df_from_multiple_files(file_paths, format=format)
    else:
        df_all = _df_from_single_file(file_paths, format=format)

    # Get maps to set as dataset attributes
    map_image_id_to_filename, map_category_to_str = (
        _get_map_attributes_from_df(df_all)
    )

    # Convert dataframe to xarray dataset
    ds = _df_to_xarray_ds(df_all)

    # Add attributes to the xarray dataset
    ds.attrs = {
        "annotation_files": file_paths,
        "annotation_format": format,
        "images_directories": images_dirs,
        "map_category_to_str": map_category_to_str,
        "map_image_id_to_filename": map_image_id_to_filename,
    }

    return ds


def _get_map_attributes_from_df(
    df: DataFrame[ValidBboxAnnotationsDataFrame],
) -> tuple[dict, dict]:
    """Get the map attributes from the dataframe.

    Parameters
    ----------
    df : DataFrame[ValidBboxesDataFrame]
        Bounding box annotations dataframe.

    Returns
    -------
    tuple[dict, dict]
        Map from "image_id" to image_filename and
        map from "category_id" to category name if present.

    """
    # Compute dataset attributes from a valid intermediate dataframe
    # map from "image_id" to image_filename
    mapping_df = df[["image_filename", "image_id"]].drop_duplicates()
    map_image_id_to_filename = mapping_df.set_index("image_id").to_dict()[
        "image_filename"
    ]
    # map from "category_id" to category name
    map_category_to_str = {}
    if all(col in df.columns for col in ["category_id", "category"]):
        map_category_to_str = (
            df[["category_id", "category"]]
            .drop_duplicates()
            .set_index("category_id")
            .to_dict()["category"]
        )

        # sort by category_id
        map_category_to_str = dict(sorted(map_category_to_str.items()))

    return (map_image_id_to_filename, map_category_to_str)


@pa.check_types
def _df_from_multiple_files(
    list_filepaths: list[Path | str], format: Literal["VIA", "COCO"]
) -> DataFrame[ValidBboxAnnotationsDataFrame]:
    """Read annotations from multiple files as a valid intermediate dataframe.

    Parameters
    ----------
    list_filepaths : list[Path | str]
        List of paths to the input annotation bounding boxes files
    format : Literal["VIA", "COCO"]
        Format of the input annotation bounding boxes files.
        Currently supported formats are "VIA" and "COCO".

    Returns
    -------
    DataFrame[ValidBboxesDataFrame]
        Intermediate dataframe for bounding boxes annotations. The dataframe is
        indexed by "annotation_id" and has the following columns:
        "image_filename", "image_id", "image_width", "image_height", "x_min",
        "y_min", "width", "height", "supercategory", "category", "category_id".

    """
    # Get list of dataframes
    df_list = [
        _df_from_single_file(file_path=file, format=format)
        for file in list_filepaths
    ]

    # Concatenate and reindex
    # the resulting axis is labeled 0,1,…,n - 1.
    # NOTE: after ignore_index=True the index name is no longer "annotation_id"
    df_all = pd.concat(df_list, ignore_index=True)

    # Update "image_id" based on the alphabetically sorted list of unique image
    # filenames across all input files
    list_image_filenames = sorted(list(df_all["image_filename"].unique()))
    df_all["image_id"] = df_all["image_filename"].apply(
        lambda x: list_image_filenames.index(x)
    )

    # Sort by image_filename
    df_all = df_all.sort_values(by=["image_filename"])

    # Remove duplicates that may exist across files and reindex
    # NOTE: we exclude image_width and image_height from the set of columns
    # to identify duplicates, as these may differ across files.
    df_all = df_all.drop_duplicates(
        subset=[
            col
            for col in df_all.columns
            if col not in ["image_width", "image_height"]
        ],
        ignore_index=True,
        inplace=False,
    )

    # Set the index name back to "annotation_id"
    df_all.index.name = "annotation_id"

    return df_all


@pa.check_types
def _df_from_single_file(
    file_path: Path | str, format: Literal["VIA", "COCO"]
) -> DataFrame[ValidBboxAnnotationsDataFrame]:
    """Read annotations from a single file as a valid intermediate dataframe.

    Parameters
    ----------
    file_path : Path | str
        Path to the input annotation bounding boxes file.
    format : Literal["VIA", "COCO"]
        Format of the input bounding boxes annotation file.
        Currently supported formats are "VIA" and "COCO".

    Returns
    -------
    DataFrame[ValidBboxesDataFrame]
        Intermediate dataframe for bounding boxes annotations. The dataframe
        is indexed by "annotation_id" and has the following columns:
        'image_filename', 'image_id', 'image_width', 'image_height',
        'x_min', 'y_min', 'width', 'height', 'supercategory', 'category',
        'category_id'.

    """
    # Choose the appropriate validator and row-extraction function
    validator: type[ValidVIA | ValidCOCO]
    if format == "VIA":
        validator = ValidVIA
        get_rows_from_file = _df_rows_from_valid_VIA_file
    elif format == "COCO":
        validator = ValidCOCO
        get_rows_from_file = _df_rows_from_valid_COCO_file
    else:
        raise ValueError(f"Unsupported format: {format}")

    # Build dataframe from extracted rows
    valid_file = validator(file_path)
    list_rows = get_rows_from_file(valid_file.path)
    df = pd.DataFrame(list_rows)

    # Sort annotations by image_filename
    df = df.sort_values(by=["image_filename"])

    # Drop duplicates and reindex
    # The resulting axis is labeled 0,1,…,n-1.
    df = df.drop_duplicates(
        subset=[col for col in df.columns if col != "annotation_id"],
        ignore_index=True,
        inplace=False,
    )

    # Cast bbox coordinates and shape as floats
    for col in ["x_min", "y_min", "width", "height"]:
        df[col] = df[col].astype(np.float64)

    # Set the index name to "annotation_id"
    df = df.set_index("annotation_id")

    return df


def _df_rows_from_valid_VIA_file(file_path: Path) -> list[dict]:
    """Extract list of dataframe rows from a validated VIA JSON file.

    Parameters
    ----------
    file_path : Path
        Path to the validated VIA JSON file.

    Returns
    -------
    list[dict]
        List of dataframe rows extracted from the validated VIA JSON file.

    """
    # Read validated json as dict
    with open(file_path) as file:
        data_dict = json.load(file)

    # Get list of sorted image filenames
    image_metadata_dict = data_dict["_via_img_metadata"]
    list_sorted_filenames = sorted(
        [img_dict["filename"] for img_dict in image_metadata_dict.values()]
    )

    # Get supercategories and categories
    via_attributes = data_dict["_via_attributes"]
    supercategories_dict = {}
    if "region" in via_attributes:
        supercategories_dict = via_attributes["region"]

    # Compute list of rows in dataframe
    list_rows = []
    annotation_id = 0
    # loop through images
    for _, img_dict in image_metadata_dict.items():
        # Extract img width and height,
        # set to default if invalid or not present
        image_width, image_height = (
            _get_image_shape_attr_as_integer(
                img_dict["file_attributes"],
                file_attr,  # type: ignore
            )
            for file_attr in ["width", "height"]
        )

        # loop thru annotations in the image
        for region in img_dict["regions"]:
            # Extract region data
            region_shape = region["shape_attributes"]
            region_attributes = region["region_attributes"]

            # Extract category data if present
            if region_attributes and supercategories_dict:
                # supercategory
                # A region (bbox) can have multiple supercategories.
                # We only consider the first supercategory in alphabetical
                # order.
                supercategory = sorted(list(region_attributes.keys()))[0]

                # category name
                # in VIA files, the category_id is a string
                category_id_str = region_attributes[supercategory]
                categories_dict = supercategories_dict[supercategory][
                    "options"
                ]
                category = categories_dict[category_id_str]

                # category_id as int
                category_id = _category_id_as_int(
                    category_id_str, categories_dict
                )

            else:
                supercategory, category, category_id = (
                    ValidBboxAnnotationsDataFrame.get_empty_values()[key]
                    for key in ["supercategory", "category", "category_id"]
                )

            # Add to row
            row = {
                "annotation_id": annotation_id,
                "image_filename": img_dict["filename"],
                "image_id": list_sorted_filenames.index(img_dict["filename"]),
                "image_width": image_width,
                "image_height": image_height,
                "x_min": region_shape["x"],
                "y_min": region_shape["y"],
                "width": region_shape["width"],
                "height": region_shape["height"],
                "supercategory": supercategory,
                "category": category,
                "category_id": category_id,
            }

            list_rows.append(row)

            # update "annotation_id"
            annotation_id += 1

    return list_rows


def _get_image_shape_attr_as_integer(
    file_attrs: dict, attr_name: Literal["width", "height"]
) -> int:
    """Safely extract the image shape attribute as an integer.

    If the attribute is not present or invalid, return the default value for
    the image shape attribute defined in
    ValidBboxesDataFrame.get_empty_values().

    The file_attrs dictionary should come from a VIA input file.

    Parameters
    ----------
    file_attrs : dict
        File attributes dictionary extracted from a VIA input file.
    attr_name : Literal["width", "height"]
        Name of the image shape attribute.

    Returns
    -------
    int
        Attribute value as int. If the attribute is not present or invalid,
        return the default value for the image shape attribute defined in
        ValidBboxesDataFrame.get_empty_values().

    """
    default_value = ValidBboxAnnotationsDataFrame.get_empty_values()[
        f"image_{attr_name}"
    ]
    try:
        return int(file_attrs.get(attr_name, default_value))
    except (TypeError, ValueError):
        return default_value


def _category_id_as_int(
    category_id_str: str, list_categories: list[str]
) -> int:
    """Convert category_id to int if possible, otherwise factorize it.

    The category_id is a string in VIA files. If it cannot be converted to an
    integer, it is factorized to a 1-based integer (0 is reserved for the
    background class) based on the alphabetically sorted list of categories.

    Parameters
    ----------
    category_id_str : str
        Category ID as string.
    list_categories : list[str]
        List of categories.

    Returns
    -------
    int
        Category ID as int.

    """
    # get category_id as int
    try:
        category_id = int(category_id_str)
    except ValueError:
        # factorize to 0-based integers
        list_sorted_options = sorted(list_categories)
        category_id = list_sorted_options.index(category_id_str)
        # Add 1 to the factorised values to make them 1-based
        category_id = category_id + 1

    return category_id


def _df_rows_from_valid_COCO_file(file_path: Path) -> list[dict]:
    """Extract list of dataframe rows from a validated COCO JSON file.

    Parameters
    ----------
    file_path : Path
        Path to the validated COCO JSON file.

    Returns
    -------
    list[dict]
        List of dataframe rows extracted from the validated COCO JSON file.

    """
    # Read validated json as dict
    with open(file_path) as file:
        data_dict = json.load(file)

    # Prepare data
    # We define "image_id_ethology" as the 0-based index of the image
    # following the alphabetically sorted list of unique image filenames.
    # In the following we assume the list of images under "images" in the
    # COCO JSON file is unique (i.e. it has no duplicate elements).
    map_img_id_coco_to_ethology = {
        img_dict["id"]: idx
        for idx, img_dict in enumerate(
            sorted(data_dict["images"], key=lambda x: x["file_name"])
        )
    }
    map_img_id_coco_to_filename = {
        img_dict["id"]: img_dict["file_name"]
        for img_dict in data_dict["images"]
    }
    map_img_id_coco_to_width_height = {
        img_dict["id"]: (img_dict["width"], img_dict["height"])
        for img_dict in data_dict["images"]
    }  # COCO files from VGG annotator always have
    # image width and height (can be 0)
    map_category_id_to_category_data = {
        cat_dict["id"]: (cat_dict["name"], cat_dict.get("supercategory", ""))
        for cat_dict in data_dict["categories"]
    }  # category data: category name, supercategory name

    # Build standard dataframe
    list_rows = []
    for annot_id, annot_dict in enumerate(data_dict["annotations"]):
        # image data
        img_id_coco = annot_dict["image_id"]
        image_filename = map_img_id_coco_to_filename[img_id_coco]
        image_width, image_height = map_img_id_coco_to_width_height[
            img_id_coco
        ]

        # compute image ID following ethology convention
        img_id_ethology = map_img_id_coco_to_ethology[img_id_coco]

        # bbox data
        x_min, y_min, width, height = annot_dict["bbox"]

        # category data
        category_id = annot_dict["category_id"]
        category, supercategory = map_category_id_to_category_data[category_id]

        row = {
            "annotation_id": annot_id,
            "image_filename": image_filename,
            "image_id": img_id_ethology,
            "image_width": image_width,
            "image_height": image_height,
            "x_min": x_min,
            "y_min": y_min,
            "width": width,
            "height": height,
            "supercategory": supercategory,  # if not defined, set to ""
            "category": category,
            "category_id": category_id,
            # in COCO files, the category_id is always a 1-based integer
        }

        list_rows.append(row)

    return list_rows


@pa.check_types
def _df_to_xarray_ds(
    df: DataFrame[ValidBboxAnnotationsDataFrame],
) -> xr.Dataset:
    """Convert a bounding box annotations dataframe to an xarray dataset.

    Parameters
    ----------
    df : DataFrame[ValidBboxesDataFrame]
        A valid intermediate dataframe for bounding boxes annotations.

    Returns
    -------
    xr.Dataset
        an xarray dataset with the following dimensions:
        - `image_id`: holds the 0-based index of the image in the "images"
        list of the COCO JSON file;
        - `space`: `x` or `y`;
        - `id`: annotation ID per image, assigned from 0 to the max number of
        annotations per image in the full dataset. Note that the annotation IDs
        are not necessarily consistent across images. This means that the
        annotations with ID `m` in image `t` and image `t+1` will likely not
        correspond to the same individual.

        The dataset is made up of the following arrays:
        - `position`: (`image_id`, `space`, `id`)
        - `shape`: (`image_id`, `space`, `id`)
        - `category`: (`image_id`, `id`)

    """
    # Drop columns if all values in that column are empty
    default_values = ValidBboxAnnotationsDataFrame.get_empty_values()
    list_empty_cols = [
        col for col in default_values if all(df[col] == default_values[col])
    ]
    df = df.drop(columns=list_empty_cols)

    # Compute max number of annotations per image
    max_annotations_per_image = df["image_id"].value_counts().max()

    # Sort the dataframe by image_id
    df = df.sort_values(by=["image_id"])

    # Compute indices of the rows where the image ID switches
    bool_id_diff_from_prev = df["image_id"].ne(df["image_id"].shift())
    indices_id_switch = np.argwhere(bool_id_diff_from_prev)[1:, 0]

    # Extract arrays from the dataframe
    arrays_metadata = _prepare_array_dicts(df)
    array_dict = _extract_arrays_from_df(
        df, arrays_metadata, indices_id_switch, max_annotations_per_image
    )

    # Build data vars dictionary
    data_vars = {
        array_key.split("_array")[0]: (
            arrays_metadata[array_key]["dims"],
            array_dict[array_key],
        )
        for array_key in array_dict
    }

    return xr.Dataset(
        data_vars=data_vars,
        coords=dict(
            image_id=df["image_id"].unique(),
            space=["x", "y"],
            id=range(max_annotations_per_image),
        ),
    )


def _prepare_array_dicts(
    df: pd.DataFrame,
) -> dict[str, dict[str, Any]]:
    """Prepare the metadata for the arrays in the xarray dataset.

    Parameters
    ----------
    df : pd.DataFrame
        A dataframe for bounding boxes annotations.

    Returns
    -------
    dict[str, dict[str, Any]]
        A dictionary with the metadata for the arrays in the xarray dataset.

    """
    arrays_metadata: dict[str, dict[str, Any]] = {
        "position_array": {
            "columns": ["x_min", "y_min"],
            "type": np.float64,
            "pad_value": np.nan,
            "dims": ("image_id", "space", "id"),
        },
        "shape_array": {
            "columns": ["width", "height"],
            "type": np.float64,
            "pad_value": np.nan,
            "dims": ("image_id", "space", "id"),
        },
    }

    # Add image shape data if present
    if all(col in df.columns for col in ["image_width", "image_height"]):
        arrays_metadata["image_shape_array"] = {
            "columns": ["image_width", "image_height"],
            "type": int,
            "pad_value": -1,
            "dims": ("image_id", "space"),
        }

    # Add category data if present
    if all(col in df.columns for col in ["category_id", "category"]):
        arrays_metadata["category_array"] = {
            "columns": ["category_id"],
            "type": int,
            "pad_value": -1,
            "dims": ("image_id", "id"),
        }

    return arrays_metadata


def _extract_arrays_from_df(
    df: pd.DataFrame,
    arrays_metadata: dict[str, dict[str, Any]],
    indices_id_switch: np.ndarray,
    max_annotations_per_image: int,
) -> dict[str, np.ndarray]:
    """Extract arrays in metadata dict from a df of bounding boxes annotations.

    Parameters
    ----------
    df : pd.DataFrame
        A dataframe for bounding boxes annotations.
    arrays_metadata : dict[str, dict[str, Any]]
        A dictionary with the metadata for the arrays to extract.
    indices_id_switch : np.ndarray
        Indices of the rows where the image ID switches.
    max_annotations_per_image : int
        The maximum number of annotations per image.

    Returns
    -------
    dict[str, np.ndarray]
        A dictionary with the arrays extracted from the dataframe.

    """
    array_dict = {}
    for key in arrays_metadata:
        # Extract annotations per image
        list_arrays = np.split(
            df[arrays_metadata[key]["columns"]].to_numpy(
                dtype=arrays_metadata[key]["type"]
            ),
            indices_id_switch,
        )  # each array: (n_annotations, N_DIM)

        if key == "image_shape_array":
            array_dict[key] = np.stack(
                [np.unique(arr, axis=0) for arr in list_arrays], axis=0
            ).squeeze(axis=1)  # (n_images, N_DIM)

        else:
            # Pad arrays with NaN values along the annotation ID axis
            # and stack to (n_images, n_max_annotations, N_DIM)
            list_arrays_padded = [
                np.pad(
                    arr,
                    ((0, max_annotations_per_image - arr.shape[0]), (0, 0)),
                    constant_values=arrays_metadata[key]["pad_value"],
                )
                for arr in list_arrays
            ]
            array_dict[key] = np.stack(list_arrays_padded, axis=0)

            # Reorder dimensions to (n_images, N_DIM, n_max_annotations)
            # (squeeze the N_DIM axis (N_DIM=1) for "category")
            array_dict[key] = np.moveaxis(array_dict[key], -1, 1)
            if key == "category_array":
                array_dict[key] = array_dict[key].squeeze(axis=1)

    # Modify x_min and y_min to represent the bbox centre
    array_dict["position_array"] += array_dict["shape_array"] / 2

    return array_dict



"""Sphinx configuration for ethology documentation."""

import os
import sys
from importlib.metadata import version as get_version

from sphinx_gallery import sorting

# Used when building API docs, put the dependencies
# of any class you are documenting here
autodoc_mock_imports: list[str] = ["cv2"]

# Add the module path to sys.path here.
# If the directory is relative to the documentation root,
# use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("../.."))

project = "ethology"
copyright = "2024, University College London"
author = "Adam Tyson"
try:
    release = get_version("ethology")
    release = release.split("+")[0]  # remove git hash
except LookupError:
    # if git is not initialised, still allow local build
    # with a dummy version
    release = "0.0.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "myst_parser",
    "nbsphinx",
    "notfound.extension",
    "sphinx_design",
    "sphinx_gallery.gen_gallery",
    "sphinx_sitemap",
    "sphinx.ext.autosectionlabel",
]

# Configure the myst parser to enable cool markdown features
# See https://sphinx-design.readthedocs.io
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "tasklist",
]
# Automatically add anchors to markdown headings
myst_heading_anchors = 4

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Automatically generate stub pages for API
autosummary_generate = True
autosummary_generate_overwrite = False
autodoc_default_options = {"show-inheritance": True}  # applies to all classes

# Prefix section labels with the document name
autosectionlabel_prefix_document = True


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "**.ipynb_checkpoints",
    # to ensure that include files (partial pages) aren't built, exclude them
    # https://github.com/sphinx-doc/sphinx/issues/1965#issuecomment-124732907
    "**/includes/**",
    # exclude .py and .ipynb files in examples generated by sphinx-gallery
    # this is to prevent sphinx from complaining about duplicate source files
    "examples/*.ipynb",
    "examples/*.py",
    "examples/**/*.ipynb",
    "examples/**/*.py",
]

# -- Options for HTML output -------------------------------------------------
html_theme = "pydata_sphinx_theme"
html_title = "ethology"

# Customize the theme
html_theme_options = {
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/neuroinformatics-unit/ethology",
            # Icon class (if "type": "fontawesome"),
            # or path to local image (if "type": "local")
            "icon": "fa-brands fa-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        },
        {
            "name": "Zulip (chat)",
            "url": "https://neuroinformatics.zulipchat.com/#narrow/channel/483869-Ethology",
            "icon": "fa-solid fa-comments",
            "type": "fontawesome",
        },
    ],
    "logo": {
        "text": f"{project} v{release}",
    },
    "footer_start": ["footer_start"],
    "footer_end": ["footer_end"],
    "external_links": [],
}

# Redirect the webpage to another URL
# Sphinx will create the appropriate CNAME file in the build directory
# The default is the URL of the GitHub pages
# https://www.sphinx-doc.org/en/master/usage/extensions/githubpages.html
github_user = "neuroinformatics-unit"
html_baseurl = "https://ethology.neuroinformatics.dev"
sitemap_url_scheme = "{link}"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = [
    ("css/custom.css", {"priority": 100}),
]
html_favicon = "_static/light-logo-niu.png"

# The linkcheck builder will skip verifying that anchors exist when checking
# these URLs
linkcheck_anchors_ignore_for_url = [
    "https://neuroinformatics.zulipchat.com/",
    "https://gin.g-node.org/G-Node/Info/wiki/",
    "https://cocodataset.org/",
]
# A list of regular expressions that match URIs that should not be checked
linkcheck_ignore = [
    "https://opensource.org/license/bsd-3-clause/",  # to avoid odd 403 error
]

myst_url_schemes = {
    "http": None,
    "https": None,
    "ftp": None,
    "mailto": None,
    "ethology-github": "https://github.com/neuroinformatics-unit/ethology/{{path}}",
    "ethology-zulip": "https://neuroinformatics.zulipchat.com/#narrow/channel/483869-Ethology",
    "setuptools-scm": "https://setuptools-scm.readthedocs.io/en/latest/{{path}}#{{fragment}}",
    "sphinx-doc": "https://www.sphinx-doc.org/en/master/usage/{{path}}#{{fragment}}",
    "sphinx-gallery": "https://sphinx-gallery.github.io/stable/{{path}}#{{fragment}}",
    "conda": "https://docs.conda.io/en/latest/",
    "gin": "https://gin.g-node.org/{{path}}#{{fragment}}",
    "github-docs": "https://docs.github.com/en/{{path}}#{{fragment}}",
    "mamba": "https://mamba.readthedocs.io/en/latest/",
    "myst-parser": "https://myst-parser.readthedocs.io/en/latest/{{path}}#{{fragment}}",
    "napari": "https://napari.org/dev/{{path}}",
}

intersphinx_mapping = {
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "python": ("https://docs.python.org/3", None),
    "loguru": ("https://loguru.readthedocs.io/en/stable/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandera": ("https://pandera.readthedocs.io/en/stable/", None),
    "movement": ("https://movement.neuroinformatics.dev/latest/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "jsonschema": (
        "https://python-jsonschema.readthedocs.io/en/stable/",
        None,
    ),
}


# What to show on the 404 page
notfound_context = {
    "title": "Page Not Found",
    "body": """
<h1>Page Not Found</h1>

<p>Sorry, we couldn't find that page.</p>

<p>Try using the search box or go to the homepage.</p>

<p>You can also email us on code@adamltyson.com.</p>

""",
}

# needed for GH pages (vs readthedocs),
# because we have no '/<language>/<version>/' in the URL
notfound_urls_prefix = None


# sphinx-gallery configuration

sphinx_gallery_conf = {
    "examples_dirs": ["../../examples"],
    "within_subsection_order": sorting.ExplicitOrder(
        [
            "coco_bbox_ethology_and_movement.py",
            "approximate_subset_sum_split.py",
            "*",
        ]
    ),
    "filename_pattern": "/*.py",  # which files to execute before inclusion
    "gallery_dirs": ["examples"],  # output directory
    "run_stale_examples": True,  # re-run examples on each build
    # Integration with Binder, see https://sphinx-gallery.github.io/stable/configuration.html#generate-binder-links-for-gallery-notebooks-experimental
    "binder": {
        "org": "neuroinformatics-unit",
        "repo": "ethology",
        "branch": "gh-pages",
        "binderhub_url": "https://mybinder.org",
        "dependencies": ["environment.yml"],
    },
    "reference_url": {"ethology": None},
    "default_thumb_file": "source/_static/dark-logo-niu.png",
    # for examples card
    "remove_config_comments": True,
    # Do not render comments with the pattern # sphinx_gallery_config [= value]
    # Mini-galleries config
    # see https://sphinx-gallery.github.io/stable/configuration.html#add-mini-galleries-for-api-documentation
    "backreferences_dir": "api/backreferences",
    # directory where function/class granular galleries are stored
    "doc_module": ("ethology",),
    # module for which to generate mini-galleries
    "promote_jupyter_magic": True,
    # any code block starting with Jupyter cell magics
    # (e.g. %%bash or %%writefile) will be turned into a runnable code block.
}



"""Utilities for splitting annotations datasets."""

from collections import Counter
from typing import TypedDict

import numpy as np
import xarray as xr
from loguru import logger
from sklearn.model_selection import GroupKFold


def split_dataset_group_by(
    dataset: xr.Dataset,
    group_by_var: str,
    list_fractions: list[float],
    samples_coordinate: str = "image_id",
    method: str = "auto",
    seed: int = 42,
    epsilon: float = 0,
) -> tuple[xr.Dataset, xr.Dataset]:
    """Split an annotations dataset by grouping variable.

    Split an ``ethology`` annotations dataset into two subsets ensuring that
    the subsets are disjoint in the grouping variable (i.e., no group appears
    in both subsets). The function automatically chooses between a
    "group k-fold" approach and an "approximate subset-sum" approach based on
    the number of unique groups and requested split fractions.

    Parameters
    ----------
    dataset : xarray.Dataset
        The annotations dataset to split.
    group_by_var : str
        The grouping variable to use for splitting the dataset. Must be
        1-dimensional along the ``samples_coordinate``.
    list_fractions : list[float, float]
        The fractions of the input annotations dataset to allocate to
        each subset. Must contain only two elements and sum to 1.
    samples_coordinate : str, optional
        The coordinate along which to split the dataset. Default is
        ``image_id``.
    method : str, optional
        Method to use: ``auto``, ``kfold``, or ``apss``. When ``auto``,
        it automatically selects between ``kfold`` or ``apss`` based on
        the number of unique groups. See Notes for further details.
        Default is ``auto``.
    seed : int, optional
        Random seed for reproducibility used in the "group k-fold" approach.
        Controls both the shuffling of the sample indices and the random
        selection of the output split from all possible ones. Only used when
        ``method`` is ``kfold`` or when ``auto`` selects ``kfold``.
        Default is 42.
    epsilon : float, optional
        The approximation tolerance for the "approximate subset-sum" approach,
        expressed as a fraction of 1. The sum of samples in the smallest subset
        is guaranteed to be at least ``(1 - epsilon)`` times the optimal sum
        for the requested fractions and grouping variable. When ``epsilon``
        is 0, the algorithm finds the exact optimal sum. Larger values result
        in faster computation but may yield subsets with a total number of
        samples further from the optimal. Only used when ``method`` is
        ``apss`` or when ``auto`` selects ``apss``. Default is 0.

    Returns
    -------
    tuple[xarray.Dataset, xarray.Dataset]
        The two subsets of the input dataset. The subsets are returned in the
        same order as the input list of fractions ``list_fractions``.

    Raises
    ------
    ValueError
        If the elements of ``list_fractions`` are not exactly two, are not
        between 0 and 1, or do not sum to 1. If ``group_by_var`` is not
        1-dimensional along the ``samples_coordinate``. If ``method`` is
        ``kfold`` but there are insufficient groups for the requested split
        fractions.

    Notes
    -----
    When ``method`` is ``auto``, the function automatically selects between two
    approaches:

    - **Group k-fold method** (default when sufficient groups exist): used
      when the number of unique groups is greater than or equal to the number
      of required folds (calculated as ``1 / min(list_fractions)``). This
      method computes all possible partitions of groups into folds and randomly
      selects one of them as the output split. The selection is controlled by
      the ``seed`` parameter for reproducibility. We use
      :class:`sklearn.model_selection.GroupKFold` cross-validator to compute
      all possible partitions of groups into folds.

    - **Approximate subset-sum method** (fallback): used when there are too few
      unique groups for group k-fold splitting. This method deterministically
      finds a subset of groups whose combined sample count best matches the
      requested fractions. The ``epsilon`` parameter controls the
      speed-accuracy tradeoff. When ``epsilon`` is 0, the algorithm finds the
      exact optimal sum. Larger values of ``epsilon`` result in faster
      computation but may yield subsets with a total number of samples further
      from the optimal. In cases where no valid split exists (e.g., all groups
      have more samples than the target), one subset may be empty and a
      warning is logged.

    See Also
    --------
    :class:`sklearn.model_selection.GroupKFold` : Group k-fold cross-validator.

    Examples
    --------
    Split a dataset of 100 images extracted from 10 different videos.
    The xarray dataset has a single data variable ``video_id`` defined along
    the ``image_id`` dimension. We would like to compute an 80/20 split,
    ensuring the subsets of the dataset are disjoint in the grouping
    variable ``video_id``. Since there are many unique groups
    (i.e., unique video IDs), the function automatically selects
    the "group k-fold" method.

    >>> import xarray as xr
    >>> from ethology.datasets.split import split_dataset_group_by
    >>> ds_large = xr.Dataset(
    ...     data_vars=dict(
    ...         video_id=("image_id", np.tile(np.arange(10), 10)),
    ...     ),  # 10 different video IDs across 100 images
    ...     coords=dict(image_id=range(100)),
    ... )
    >>> ds_subset_1, ds_subset_2 = split_dataset_group_by(
    ...     ds_large, "video_id", [0.8, 0.2], seed=42
    ... )
    >>> print(len(ds_subset_1.image_id) / len(ds_large.image_id))  # 0.8
    >>> print(len(ds_subset_2.image_id) / len(ds_large.image_id))  # 0.2


    Using different seeds produces different splits when the "group k-fold"
    method is used:

    >>> ds_subset_1b, ds_subset_2b = split_dataset_group_by(
    ...     ds_large, "video_id", [0.8, 0.2], seed=123
    ... )
    >>> assert not ds_subset_1.equals(ds_subset_1b)
    >>> assert not ds_subset_2.equals(ds_subset_2b)
    >>> print(len(ds_subset_1b.image_id) / len(ds_large.image_id))  # 0.8
    >>> print(len(ds_subset_2b.image_id) / len(ds_large.image_id))  # 0.2

    The function automatically selects the appropriate method. In the example
    below, a smaller dataset with 3 unique video IDs is used. With a 0.2
    minimum fraction (requiring 5 folds), the "group k-fold" method cannot be
    used, since there would be more folds than groups. Therefore, the
    "approximate subset-sum" method is selected automatically and an
    approximate split is returned. Note that when using the ``apss`` method,
    the seed value is ignored.

    >>> ds_small = xr.Dataset(
    ...     data_vars=dict(
    ...         video_id=("image_id", [1, 0, 0, 2, 2, 2, 0, 0, 0, 0, 0, 1]),
    ...     ),
    ...     coords=dict(image_id=range(12)),
    ... )
    >>> ds_subset_1, ds_subset_2 = split_dataset_group_by(
    ...     ds_small,
    ...     group_by_var="video_id",
    ...     list_fractions=[0.8, 0.2],
    ... )
    >>> print(len(ds_subset_1.image_id) / len(ds_small.image_id))  # 0.833
    >>> print(len(ds_subset_2.image_id) / len(ds_small.image_id))  # 0.166

    The ``epsilon`` parameter controls the approximation for the subset-sum
    method when auto-selected or explicitly specified:

    >>> ds_subset_1, ds_subset_2 = split_dataset_group_by(
    ...     ds_small,
    ...     group_by_var="video_id",
    ...     list_fractions=[0.8, 0.2],
    ...     epsilon=0.1,  # accept a solution >= 90% of the optimal
    ...     method="apss",
    ... )
    >>> print(len(ds_subset_1.image_id) / len(ds_small.image_id))  # 0.833
    >>> print(len(ds_subset_2.image_id) / len(ds_small.image_id))  # 0.166

    """
    # Checks
    if sum(list_fractions) != 1:
        raise ValueError("The split fractions must sum to 1.")
    if len(list_fractions) != 2:
        raise ValueError("The list of fractions must have only two elements.")
    if any(fraction < 0 or fraction > 1 for fraction in list_fractions):
        raise ValueError("The split fractions must be between 0 and 1.")
    if len(dataset[group_by_var].shape) != 1:
        raise ValueError(
            f"The grouping variable {group_by_var} must be 1-dimensional along"
            f" {samples_coordinate}."
        )

    # Count unique groups
    n_unique_groups = len(np.unique(dataset[group_by_var].values))
    n_required_folds = int(np.rint(1 / min(list_fractions)))

    # Auto-select method
    if method == "auto":
        if n_unique_groups >= n_required_folds:
            method = "kfold"
        else:
            method = "apss"
            logger.info(
                f"Only {n_unique_groups} unique groups exist but "
                f"{n_required_folds} are required for k-fold method. "
                "Auto-selected approximate subset-sum method "
                f"with epsilon={epsilon}. Seed setting is ignored."
            )

    # Dispatch to appropriate method
    if method == "kfold":
        logger.info(
            f"Using group k-fold method with {n_required_folds} folds "
            f"and seed={seed}."
        )
        return _split_dataset_group_by_kfold(
            dataset, group_by_var, list_fractions, samples_coordinate, seed
        )
    elif method == "apss":
        logger.info(
            f"Using approximate subset-sum method with epsilon={epsilon}."
        )
        return _split_dataset_group_by_apss(
            dataset, group_by_var, list_fractions, epsilon, samples_coordinate
        )
    else:
        raise ValueError(f"Unknown method: {method}")


def _split_dataset_group_by_kfold(
    dataset: xr.Dataset,
    group_by_var: str,
    list_fractions: list[float],
    samples_coordinate: str = "image_id",
    seed: int = 42,
) -> tuple[xr.Dataset, xr.Dataset]:
    """Split an annotations dataset using scikit-learn's GroupKFold.

    Split an ``ethology`` annotations dataset into two subsets
    ensuring that the subsets are disjoint in the grouping variable.
    This method uses scikit-learn's GroupKFold cross-validator to
    randomly partition groups into folds and then selects one fold
    as the smaller subset and the remaining folds as the larger subset.

    Parameters
    ----------
    dataset : xarray.Dataset
        The annotations dataset to split.
    group_by_var : str
        The grouping variable to use for splitting the dataset. Must be
        1-dimensional along the ``samples_coordinate``.
    list_fractions : list[float, float]
        The fractions of the input annotations dataset to allocate to
        each subset. Must contain only two elements and sum to 1.
    samples_coordinate : str, optional
        The coordinate along which to split the dataset. Default is
        ``image_id``.
    seed : int, optional
        Random seed for reproducibility. Controls both the GroupKFold
        indices shuffling and the random selection of the output split from
        all the possible ones. Default is 42.

    Returns
    -------
    tuple[xarray.Dataset, xarray.Dataset]
        The two subsets of the input dataset. The subsets are returned in the
        same order as the input list of fractions.

    Raises
    ------
    ValueError

        If the elements of ``list_fractions`` are not exactly two, are not
        between 0 and 1, or do not sum to 1. If ``group_by_var`` is not
        1-dimensional along the ``samples_coordinate``.

    Examples
    --------
    Split a dataset with a single data variable ``foo`` defined along the
    ``image_id`` dimension into an 80/20 split, ensuring that the
    subsets are disjoint in the grouping variable ``foo``.

    >>> from ethology.datasets.split import _split_dataset_group_by_kfold
    >>> import xarray as xr
    >>> ds = xr.Dataset(
    >>>     data_vars=dict(
    >>>         foo=("image_id", [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]),
    >>>     ),  # 0: 10 counts, 1: 2 counts
    >>>     coords=dict(
    >>>         image_id=range(12),
    >>>     ),
    >>> )
    >>> ds_subset_1, ds_subset_2 = _split_dataset_group_by_kfold(
    >>>     ds,
    >>>     group_by_var="foo",
    >>>     list_fractions=[0.2, 0.8],
    >>>     seed=42,
    >>> )

    The ``seed`` parameter ensures reproducibility. Using the same seed
    on the same dataset will always produce the same split. Using a different
    seed will produce a different split.

    >>> ds_subset_1b, ds_subset_2b = _split_dataset_group_by_kfold(
    >>>     ds,
    >>>     group_by_var="foo",
    >>>     list_fractions=[0.2, 0.8],
    >>>     seed=123,
    >>> )
    >>> assert not ds_subset_1.equals(ds_subset_1b)
    >>> assert not ds_subset_2.equals(ds_subset_2b)

    """
    # Initialise k-fold iterator
    n_folds_per_shuffle = int(np.rint(1 / min(list_fractions)))
    gkf = GroupKFold(
        n_splits=n_folds_per_shuffle, shuffle=True, random_state=seed
    )

    # Compute all possible shuffles
    # In each shuffle, one fold is the test set,
    # the rest of folds make up the train set
    train_test_idcs_per_shuffle = list(
        gkf.split(
            dataset[samples_coordinate].values,
            groups=dataset[group_by_var].values,
        )
    )

    # Randomly pick one of the shuffles
    rng = np.random.default_rng(seed)
    shuffle_idx = rng.choice(len(train_test_idcs_per_shuffle))
    train_idcs, test_idcs = train_test_idcs_per_shuffle[shuffle_idx]

    # Split the datasets
    ds_train = dataset.isel({samples_coordinate: train_idcs})
    ds_test = dataset.isel({samples_coordinate: test_idcs})
    list_ds_sorted = [ds_test, ds_train]  # sorted in increasing size

    # Return datasets in the same order as the input list of fractions
    idcs_sorted = np.argsort(list_fractions)  # idcs to map input -> sorted
    idcs_original = np.argsort(idcs_sorted)  # idcs to map sorted -> input

    return tuple(list_ds_sorted[i] for i in idcs_original)


def _split_dataset_group_by_apss(
    dataset: xr.Dataset,
    group_by_var: str,
    list_fractions: list[float],
    epsilon: float = 0,
    samples_coordinate: str = "image_id",
) -> tuple[xr.Dataset, xr.Dataset]:
    """Split an annotations dataset using an approximate subset-sum approach.

    Split an ``ethology`` annotations dataset into two subsets
    ensuring that the subsets are disjoint in the grouping variable.

    Parameters
    ----------
    dataset : xarray.Dataset
        The annotations dataset to split.
    group_by_var : str
        The grouping variable to use for splitting the dataset. Must be
        1-dimensional along the ``samples_coordinate``.
    list_fractions : list[float, float]
        The fractions of the input annotations dataset to allocate to
        each subset. Must contain only two elements and sum to 1.
    epsilon : float, optional
        The approximation tolerance for the subset sum as a fraction of 1.
        The sum of samples in the smallest subset is guaranteed to be at
        least ``(1 - epsilon)`` times the optimal sum for the requested
        fractions and grouping variable. When ``epsilon`` is 0, the algorithm
        finds the exact optimal sum. Larger values result in faster
        computation but may yield subsets with a total number of samples
        further from optimal. Default is 0.
    samples_coordinate : str, optional
        The coordinate along which to split the dataset. Default is
        ``image_id``.

    Returns
    -------
    tuple[xarray.Dataset, xarray.Dataset]
        The two subsets of the input dataset. The subsets are returned in the
        same order as the input list of fractions.

    Raises
    ------
    ValueError

        If the elements of ``list_fractions`` are not exactly two, are not
        between 0 and 1, or do not sum to 1. If ``group_by_var`` is not
        1-dimensional along the ``samples_coordinate``.

    Examples
    --------
    Split a dataset with a single data variable ``foo`` defined along the
    ``image_id`` dimension into an approximate 80/20 split, ensuring that the
    subsets are disjoint in the grouping variable ``foo``.

    >>> from ethology.datasets.split import _split_dataset_group_by_apss
    >>> import xarray as xr
    >>> ds = xr.Dataset(
    >>>     data_vars=dict(
    >>>         foo=("image_id", [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1]),
    >>>     ),  # 0: 10 counts, 1: 2 counts
    >>>     coords=dict(
    >>>         image_id=range(12),
    >>>     ),
    >>> )
    >>> ds_subset_1, ds_subset_2 = _split_dataset_group_by_apss(
    >>>     ds,
    >>>     group_by_var="foo",
    >>>     list_fractions=[0.2, 0.8],
    >>>     epsilon=0,
    >>> )
    >>> print(len(ds_subset_1.image_id) / len(ds.image_id))  # 0.166
    >>> print(len(ds_subset_2.image_id) / len(ds.image_id))  # 0.833

    """
    # Compute number of samples in target subset
    # the target subset is the subset with the smallest fraction.
    target_subset_count = int(
        min(list_fractions) * len(dataset.get(samples_coordinate))
    )

    # Get list of (id, count) tuples
    # Count number of samples per group and sort by count in ascending order
    count_per_group_id = Counter(dataset[group_by_var].values).most_common()[
        ::-1
    ]

    # Cast group ids to integers and create mapping
    map_group_id_int_to_original = {}
    count_per_group_id_as_int = []
    for idx, (group_id, count) in enumerate(count_per_group_id):
        # try casting group ID as integer
        try:
            int_id = int(group_id)
        # if not castable: use ID from enumerate
        except (ValueError, TypeError):
            int_id = idx
        count_per_group_id_as_int.append((int_id, count))
        map_group_id_int_to_original[int_id] = group_id

    # Get group ids (as integers) for target subset
    subset_dict = _approximate_subset_sum(
        count_per_group_id_as_int,
        target_subset_count,
        epsilon=epsilon,
    )

    # Get original group IDs (they are not necessarily integers)
    subset_group_ids = [
        map_group_id_int_to_original[x] for x in subset_dict["ids"]
    ]

    # Extract datasets for target subset and not target subset
    ds_subset = dataset.isel(
        {samples_coordinate: dataset[group_by_var].isin(subset_group_ids)}
    )
    ds_not_subset = dataset.isel(
        {samples_coordinate: ~dataset[group_by_var].isin(subset_group_ids)}
    )

    # Throw warning if a subset is empty
    if any(len(ds.image_id) == 0 for ds in [ds_subset, ds_not_subset]):
        logger.warning("At least one of the subset datasets is empty.")

    # Return datasets in the same order as the input list of fractions
    # (argsort twice gives the inverse permutation)
    idcs_sorted = np.argsort(list_fractions)  # idcs to map input -> sorted
    idcs_original = np.argsort(idcs_sorted)  # idcs to map sorted -> input

    list_ds_sorted = [ds_subset, ds_not_subset]
    return tuple(list_ds_sorted[i] for i in idcs_original)


def split_dataset_random(
    dataset: xr.Dataset,
    list_fractions: list[float],
    seed: int = 42,
    samples_coordinate: str = "image_id",
) -> tuple[xr.Dataset, ...]:
    """Split an annotations dataset using random sampling.

    Split an ``ethology`` annotations dataset into multiple subsets by randomly
    shuffling all samples and then partitioning them sequentially according to
    the specified fractions.


    Parameters
    ----------
    dataset : xarray.Dataset
        The annotations dataset to split.
    list_fractions : list[float, ...]
        The fractions of the input annotations dataset to allocate to
        each subset. The list must contain at least two elements, all elements
        must be between 0 and 1, and add up to 1.
    seed : int, optional
        Seed to use for the random number generator. Default is 42.
    samples_coordinate : str, optional
        The coordinate along which to split the dataset. Default is
        ``image_id``.

    Returns
    -------
    tuple[xarray.Dataset, ...]
        The subsets of the input dataset. The subsets are returned in the
        same order as the input list of fractions.

    Raises
    ------
    ValueError
        If the elements of ``list_fractions`` are less than two, are not
        between 0 and 1, or do not sum to 1.

    Examples
    --------
    Split a dataset with a single data variable ``foo``, with 100 values
    defined along the ``image_id`` dimension into 70/20/10 splits.

    >>> from ethology.datasets.split import split_dataset_random
    >>> import numpy as np
    >>> import xarray as xr
    >>> ds = xr.Dataset(
    ...     data_vars=dict(
    ...         foo=("image_id", np.random.randint(0, 100, size=100)),
    ...     ),
    ...     coords=dict(
    ...         image_id=range(100),
    ...     ),
    ... )
    >>> ds_train, ds_val, ds_test = split_dataset_random(
    ...     ds,
    ...     list_fractions=[0.7, 0.2, 0.1],
    ...     seed=42,
    ... )
    >>> print(len(ds_train.image_id))  # 70
    >>> print(len(ds_val.image_id))  # 20
    >>> print(len(ds_test.image_id))  # 10

    Notes
    -----
    The function operates in two steps: first, it shuffles all sample indices
    along the ``samples_coordinate`` dimension using the provided random seed;
    then, it partitions the shuffled indices into contiguous blocks, one for
    each subset.

    The size of each block is determined by rounding down (floor) the product
    of the subset's fraction and the total number of samples. To ensure all
    samples are included, the last subset receives any remaining samples after
    the earlier subsets have been allocated. Due to this rounding behavior,
    the actual fraction for the last subset may differ slightly from the
    requested fraction.

    """
    # Checks
    if len(list_fractions) < 2:
        raise ValueError(
            "The list of fractions must have at least two elements."
        )

    if any(fraction < 0 or fraction > 1 for fraction in list_fractions):
        raise ValueError("The split fractions must be between 0 and 1.")

    if sum(list_fractions) != 1:
        raise ValueError("The split fractions must sum to 1.")

    # Compute number of samples for each split
    list_n_samples: list[int] = []
    n_total_samples = len(dataset.get(samples_coordinate))
    for fraction in list_fractions[:-1]:
        list_n_samples.append(int(fraction * n_total_samples))
    # append the remaining samples to the last split
    list_n_samples.append(n_total_samples - sum(list_n_samples))

    # Shuffle all indices
    rng = np.random.default_rng(seed)
    shuffled_idcs = rng.permutation(n_total_samples)

    # Extract datasets for each split
    list_ds = []
    start_idx = 0
    for n_samples in list_n_samples:
        end_idx = start_idx + n_samples
        list_ds.append(
            dataset.isel(
                {samples_coordinate: shuffled_idcs[start_idx:end_idx]}
            )
        )
        start_idx = end_idx

    # Throw warning if a subset is empty
    if any(len(ds.image_id) == 0 for ds in list_ds):
        logger.warning("At least one of the subset datasets is empty.")

    # Return subsets in the same order as the input list of fractions
    # (argsort twice gives the inverse permutation)
    idcs_sorted = np.argsort(list_fractions)  # idcs to map input -> sorted
    idcs_original = np.argsort(idcs_sorted)  # idcs to map sorted -> input
    return tuple(list_ds[i] for i in idcs_original)


class _SubsetDict(TypedDict):
    """Subset dictionary.

    Each subset dictionary is made up of a list of ``group IDs`` ("ids") and
    their total ``group count`` ("sum"). Used as a type definition for the
    approximate subset sum algorithm.

    Attributes
    ----------
    sum : int
        The total ``group count`` of the subset.
    ids : list[int]
        The list of ``group IDs`` in the subset.

    """

    sum: int
    ids: list[int]


def _approximate_subset_sum(
    list_id_counts: list[tuple[int, int]], target: int, epsilon: float
) -> _SubsetDict:
    """Find a subset of the input list whose sum is maximum but below target.

    The input is a list of pairs (``group IDs``, ``group count``). We want
    to extract a subset of elements from this list such that their total
    ``group count`` is as close as possible but does not exceed the
    ``target`` value.

    Parameters
    ----------
    list_id_counts : list[tuple[int, int]]
        The list of pairs (``group IDs``, ``group count``).
    target : int
        The target value for the total ``group count`` of the subset.
    epsilon : float
        The approximation tolerance for the subset sum as a fraction of 1.
        When ``epsilon`` is 0, the algorithm finds the optimal subset for the
        requested target value and input list. Larger values of ``epsilon``
        result in faster computation but may yield subsets with a total
        ``group count`` further from the optimal.

    Returns
    -------
    _SubsetDict
        The subset dictionary.

    Raises
    ------
    Warning
        If all groups in the input list have more samples than the target
        value. In this case, the function returns an empty subset.

    Notes
    -----
    The function uses a fully polynomial-time approximation scheme (FPTAS) to
    approximately solve this subset sum problem. When ``epsilon`` is 0, it
    finds the exact optimal subset. When ``epsilon`` > 0, it returns an
    approximate solution guaranteed to be within ``epsilon`` times the optimal
    sum from below. Using an ``epsilon`` value larger than 0 may be convenient
    in cases with a large number of subsets for faster runtime.

    The algorithm iteratively processes each element in the input list,
    maintaining a list of candidate subsets whose total ``group counts`` falls
    below the target. At each iteration, it removes near duplicate subsets
    from the list of candidates ("trimming") to prevent exponential growth,
    while ensuring the total approximation error stays within ``epsilon``.
    Two subsets are near duplicates if their total ``group count`` is
    sufficiently close.

    Note that ``epsilon`` bounds the error from below relative to the optimal
    subset sum, not the ``target``. If ``OPT`` is the best possible subset sum
    below or equal to ``target``, the result is guaranteed to be
    at least ``(1 - epsilon)*OPT``. E.g. for ``epsilon = 0.2``,
    the result is guaranteed to be at least ``0.8*OPT``.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Subset_sum_problem#Fully-polynomial_time_approximation_scheme
    .. [2] https://nerderati.com/bartering-for-beers-with-approximate-subset-sums/

    """
    # Checks
    if np.min([x[1] for x in list_id_counts]) > target:
        logger.warning(
            "All groups have more samples than the target value. "
            "Returning empty subset."
        )
        return {"sum": 0, "ids": []}

    # Initialize list of candidate subsets
    list_subsets: list[_SubsetDict] = [{"sum": 0, "ids": []}]

    # Loop thru list of (id, count) pairs
    for id, count in list_id_counts:
        # Add current (id, count) pair to each candidate subset in the list
        # if the resulting subset sum is below the target.
        list_subsets.extend(
            [
                {
                    "sum": subset["sum"] + count,
                    "ids": subset["ids"] + [id],
                }
                for subset in list_subsets
                if subset["sum"] + count <= target
            ]
        )

        # Remove near-duplicate subsets in terms of total group count ("sum")
        # At each iteration of the loop, trimming introduces a small error.
        # The algorithm runs n iterations (one per item), so errors can
        # compound. Using ``delta = epsilon / (2n)`` ensures that when errors
        # compound over iterations, the total error stays within ``epsilon``.
        list_subsets = _remove_near_duplicate_subsets(
            list_subsets,
            delta=float(epsilon) / (2 * len(list_id_counts)),
        )

    # Return the subset with highest sum but below the target
    return list_subsets[-1]


def _remove_near_duplicate_subsets(
    list_subsets: list[_SubsetDict], delta: float
) -> list[_SubsetDict]:
    """Remove near-duplicate subsets from the list in terms of their total sum.

    Keeps only subsets whose sum is sufficiently larger than the previous
    subset sum in ascending order. When two subsets have sums within
    ``delta``% of each other, retains the smaller one (which is visited first
    after sorting).

    Parameters
    ----------
    list_subsets : list[_SubsetDict]
        The list of candidate subsets. Each subset is a dictionary with a list
        of ``group IDs`` ("ids") and their total ``group count`` ("sum").
    delta : float
        If two subsets are within ``delta``% of each other, they are considered
        near duplicates and the smaller one is removed.

    Returns
    -------
    list[_SubsetDict]
        The list of subsets after trimming.

    """
    # Ensure list of subsets is sorted by total sum, in ascending order
    list_subsets = sorted(list_subsets, key=lambda x: x["sum"])

    # Keep only subsets whose sum is delta% larger than the previous one
    list_subsets_trimmed = [
        list_subsets[0]
    ]  # always retain the zero subset; [{"sum": 0, "ids": []}]
    previous_subset_sum = 0
    for subset in list_subsets[1:]:  # do not trim zero subset
        if subset["sum"] > previous_subset_sum * (1 + delta):
            list_subsets_trimmed.append(subset)
            previous_subset_sum = subset["sum"]

    return list_subsets_trimmed



"""Save ``ethology`` bounding box annotations datasets to various formats."""

import json
from datetime import datetime
from pathlib import Path
from typing import Any

import pandas as pd
import pandera.pandas as pa
import pytz
import xarray as xr
from pandera.typing.pandas import DataFrame

from ethology.validators.annotations import (
    ValidBboxAnnotationsCOCO,
    ValidBboxAnnotationsDataset,
    ValidCOCO,
)
from ethology.validators.utils import _check_input, _check_output


@_check_input(validator=ValidBboxAnnotationsDataset)
@_check_output(validator=ValidCOCO)  # check output is ethology-importable
def to_COCO_file(dataset: xr.Dataset, output_filepath: str | Path):
    """Save an ``ethology`` bounding box annotations dataset to a COCO file.

    Parameters
    ----------
    dataset : xarray.Dataset
        Bounding boxes annotations xarray dataset.
    output_filepath : str or pathlib.Path
        Path for the output COCO file.

    Returns
    -------
    str
        Path for the output COCO file.

    Examples
    --------
    Save annotations to a COCO file:

    >>> from ethology.io.annotations import save_bboxes
    >>> save_bboxes.to_COCO_file(ds, "path/to/output_file.json")

    """
    # Compute valid COCO dataframe from xarray dataset
    df = _to_COCO_exportable_df(dataset)

    # Create COCO dictionary from dataframe and export
    COCO_dict = _create_COCO_dict(df)
    with open(output_filepath, "w") as f:
        json.dump(COCO_dict, f, sort_keys=True, indent=2)

    return output_filepath


@_check_input(validator=ValidBboxAnnotationsDataset)
@pa.check_types
def _to_COCO_exportable_df(
    ds: xr.Dataset,
) -> DataFrame[ValidBboxAnnotationsCOCO]:
    """Convert dataset of bounding boxes annotations to a COCO-exportable df.

    The returned dataframe is validated using ValidBBoxesDataFrameCOCO.

    Parameters
    ----------
    ds : xr.Dataset
        A valid dataset of bounding boxes annotations.

    Returns
    -------
    df : pd.DataFrame
        A dataframe of bounding boxes annotations exportable to COCO.

    """
    # Prepare dataframe from xarray dataset
    df_raw = _get_raw_df_from_ds(ds)
    df = _add_COCO_data_to_df(df_raw, ds.attrs)

    # Select columns to keep
    cols_to_select = [
        "annotation_id",
        "image_id",
        "image_filename",
        "image_width",
        "image_height",
        "bbox",
        "area",
        "segmentation",
        "category",  # str
        "category_id",  # int
        "supercategory",
        "iscrowd",
    ]
    return df[cols_to_select]


@_check_input(validator=ValidBboxAnnotationsDataset)
def _get_raw_df_from_ds(ds: xr.Dataset) -> pd.DataFrame:
    """Get preliminary dataframe from a dataset of bounding boxes annotations.

    If the dataset has an "image_shape" array, the returned dataframe
    will have "image_shape_x" and "image_shape_y" columns. The returned
    dataframe will have a "category" column, filled with the relevant category
    values, or filled with -1 if no category array was present in the
    original dataset.

    The returned dataframe is not COCO-exportable.

    Parameters
    ----------
    ds : xr.Dataset
        A valid dataset of bounding boxes annotations.

    Returns
    -------
    df : pd.DataFrame
        A preliminary dataframe of bounding boxes annotations.

    """
    # Create dataframe from xarray dataset
    df_raw = ds.to_dataframe(dim_order=["image_id", "id", "space"])
    df_raw = df_raw.reset_index()

    # Remove rows where position or shape data is nan
    # (where at least one of the specified columns contains a NaN value.)
    df_raw = df_raw.dropna(subset=["position", "shape"])

    # Add "category" column if not present
    if "category" not in df_raw.columns:
        df_raw["category"] = -1

    # Pivot the dataframe to get position_x, position_y, shape_x, shape_y, etc.
    index_cols = ["image_id", "id", "category"]
    pivot_values = ["position", "shape"]
    if "image_shape" in df_raw.columns:
        pivot_values.append("image_shape")

    df_raw = df_raw.pivot_table(
        index=index_cols,
        columns="space",
        values=pivot_values,
    ).reset_index()

    # Flatten the columns
    df_raw.columns = [
        "_".join(col).strip() if col[1] != "" else col[0]
        for col in df_raw.columns.values
    ]

    # Reset type for image_shape columns if present
    if all(
        col in df_raw.columns for col in ["image_shape_x", "image_shape_y"]
    ):
        df_raw["image_shape_x"] = df_raw["image_shape_x"].astype(int)
        df_raw["image_shape_y"] = df_raw["image_shape_y"].astype(int)

    return df_raw


@pa.check_types
def _add_COCO_data_to_df(
    df: pd.DataFrame, ds_attrs: dict
) -> DataFrame[ValidBboxAnnotationsCOCO]:
    """Add COCO-required data to preliminary dataframe.

    The input dataframe is obtained from a dataset of bounding boxes
    annotations using ``_get_raw_df_from_ds`` and is not COCO-exportable.

    Parameters
    ----------
    df : pd.DataFrame
        Preliminary dataframe of bounding boxes annotations derived
        from a dataset of bounding boxes annotations.
    ds_attrs : dict
        Attributes of the dataset of bounding boxes annotations.

    Returns
    -------
    df : pd.DataFrame
        COCO-exportable dataframe of bounding boxes annotations.
        The dataframe has the following columns:
        'id', 'annotation_id',
        'image_filename', 'image_id', 'image_width', 'image_height',
        'position_x', 'position_y', 'shape_x', 'shape_y',
        'x_min',  'y_min', 'width', 'height',
        'bbox', 'area', 'segmentation',
        'category', 'supercategory', 'category_id', 'iscrowd'.

    Notes
    -----
    The 'id' column holds the annotation ID per frame, whereas
    the 'annotation_id' column holds the annotation ID across the
    whole dataset.

    """
    # image filename
    map_image_id_to_filename = ds_attrs["map_image_id_to_filename"]
    df["image_filename"] = df["image_id"].map(map_image_id_to_filename)

    # image width and height
    if all(col in df.columns for col in ["image_shape_x", "image_shape_y"]):
        df = df.rename(
            columns={
                "image_shape_x": "image_width",
                "image_shape_y": "image_height",
            },
        )
    else:
        df["image_width"] = ds_attrs.get("image_width", 0)
        df["image_height"] = ds_attrs.get("image_height", 0)

    # bbox
    df["x_min"] = df["position_x"] - df["shape_x"] / 2
    df["y_min"] = df["position_y"] - df["shape_y"] / 2
    df["width"] = df["shape_x"]
    df["height"] = df["shape_y"]
    df["bbox"] = df[["x_min", "y_min", "width", "height"]].values.tolist()

    df["area"] = df["width"] * df["height"]

    # segmentation as list of lists of coordinates
    # top-left -> top-right -> bottom-right -> bottom-left
    df["segmentation"] = df["bbox"].apply(
        lambda bbox: [
            [
                bbox[0],  # top-left x
                bbox[1],  # top-left y
                bbox[0] + bbox[2],  # top-right x
                bbox[1],  # top-right y
                bbox[0] + bbox[2],  # bottom-right x
                bbox[1] + bbox[3],  # bottom-right y
                bbox[0],  # bottom-left x
                bbox[1] + bbox[3],  # bottom-left y
            ]
        ]
    )

    # Rename "category" to "category_id" (in dataset it is an integer)
    # and compute "category" as string from "category_id"
    map_category_to_str = ds_attrs["map_category_to_str"]
    df.rename(columns={"category": "category_id"}, inplace=True)
    df["category"] = df["category_id"].map(map_category_to_str)

    # supercategory
    if "supercategory" not in df.columns:
        df["supercategory"] = ""

    # other
    df["iscrowd"] = 0

    # Set index name and add "annotation_id" as column
    df.index.name = "annotation_id"
    df["annotation_id"] = df.index

    # Sort by "image_filename" and remove duplicates
    df = df.sort_values(by=["image_filename"])
    df = df.loc[
        df.astype(str).drop_duplicates(ignore_index=True).index
    ]  # need to serialise lists first before dropping duplicates

    return df


@pa.check_types
def _create_COCO_dict(
    df: DataFrame[ValidBboxAnnotationsCOCO],
) -> dict:
    """Extract COCO dictionary from a COCO-exportable dataframe.

    Parameters
    ----------
    df : DataFrame[ValidBBoxesDataFrameCOCO]
        COCO exportable dataframe.

    Returns
    -------
    COCO_dict : dict
        COCO dictionary.

    """
    COCO_dict: dict[str, Any] = {}
    map_columns_to_COCO_fields = (
        ValidBboxAnnotationsCOCO.map_df_columns_to_COCO_fields()
    )
    for sections in ["images", "categories", "annotations"]:
        # Extract and rename required columns for this section
        list_required_columns = map_columns_to_COCO_fields[sections].keys()
        df_section = df[list_required_columns].copy()
        df_section = df_section.rename(
            columns=map_columns_to_COCO_fields[sections]
        )

        # Extract rows as lists of dictionaries
        if sections == "annotations":
            row_dicts = df_section.to_dict(orient="records")
        else:
            row_dicts = df_section.drop_duplicates().to_dict(orient="records")

        COCO_dict[sections] = row_dicts

    # Add info section to COCO_dict
    COCO_dict["info"] = {
        "date_created": datetime.now(pytz.utc).strftime(
            "%a %b %d %Y %H:%M:%S GMT%z"
        ),
        "description": "Bounding boxes annotations exported from ethology",
        "url": "https://github.com/neuroinformatics-unit/ethology",
    }

    return COCO_dict
